/*  *********************************************************************** */
/*  *********************************************************************** */

/*  FILE kdp.c - *SAT 1.2 */
/*  Consistency check in logic K */ 

/*  *********************************************************************** */
/*  *********************************************************************** */

#include "kdp.h"

/* *********************************************************************** */
/* The functions for testing K consistency of an assignment */
/* *********************************************************************** */

/* ****************************** Core loop: forall j /\_i alpha_i & beta_j */
vertex_t *Int_test_K_consistency(image_t *cur_image_ptr, dag_t *dag_ptr,
				 lsList *alpha_lsList_ptr, lsList *beta_lsList_ptr,
				 int flag_display)
{
  int          result;
  int          list_status;
  int          max_mem, saved_max;
  lsGen        beta_lsGen;
  lsList       *alphas_formula_lsList_ptr, *to_test_lsList_ptr;
  lsList       *new_alpha_lsList_ptr, *new_beta_lsList_ptr;
  lsGeneric    beta_lsGeneric;
  vertex_t     *beta_vertex_ptr;
  image_t      *new_image_ptr;  

  /* Memoizing */
  if (Use_memoS()) {
    /* Get rid of dependencies from the list of alphas; find out if the formula is unconsistent */
    if (Getf_memoS(GET_DEPENDENCIES)) {
      new_alpha_lsList_ptr = Test_depS(alpha_lsList_ptr, beta_lsList_ptr, &beta_vertex_ptr); 
/*        assert(lsLength(*alpha_lsList_ptr) != 0); */
      if (new_alpha_lsList_ptr == NIL(lsList)) return beta_vertex_ptr;
    } else {
      new_alpha_lsList_ptr = alpha_lsList_ptr;
    }

    /* Filter the list of betas to be tested through the memozing structure */
    new_beta_lsList_ptr = Test_memoS(new_alpha_lsList_ptr, beta_lsList_ptr);
    /* Nothing to check!! */
    if (new_beta_lsList_ptr == NIL(lsList)) return NIL(vertex_t);
  } else {
    /* Do nothing */
    new_beta_lsList_ptr = beta_lsList_ptr;
    new_alpha_lsList_ptr = alpha_lsList_ptr;
  }

  /* Monitor */
  saved_max = Get_param(Image_monitor_ptr(cur_image_ptr), MAX_SATO_MEMORY);
  Get_param(Image_monitor_ptr(cur_image_ptr), MAX_SATO_MEMORY) = 0;
  max_mem = 0;
  
  /* The conjunction of all the alphas is formed */
  alphas_formula_lsList_ptr = Make_cnf_conj(dag_ptr, new_alpha_lsList_ptr); 
  /* Now testing the conjunction of the alphas with each beta */
  result = 1;
  lsForEachItem(*new_beta_lsList_ptr, beta_lsGen, beta_lsGeneric) {
    beta_vertex_ptr = UNCAST_VERTEX_ITEM(beta_lsGeneric);
    /* KSAT(/\_i alpha_i & -beta_j) */
    /* This creates to_test_lsList_ptr */
#ifdef OPTIMIZE
    to_test_lsList_ptr = Merge_formulae_to_new(Make_cnf(dag_ptr, beta_vertex_ptr, NEG_CNF),
					       alphas_formula_lsList_ptr);
#else
    to_test_lsList_ptr = Merge_formulae_to_new(alphas_formula_lsList_ptr,
					       Make_cnf(dag_ptr, beta_vertex_ptr, NEG_CNF));
#endif
    
#ifdef TRACE_CONS
    if (flag_display) {
      Print_complex_call(new_alpha_lsList_ptr, beta_vertex_ptr, to_test_lsList_ptr);
    }
#endif
    
    /* We need a new image for SATO */
    new_image_ptr = Init_cnf_sat_rec(dag_ptr, to_test_lsList_ptr);
    /* Passing the reference to the monitor structure */
    Image_monitor_ptr(new_image_ptr) = Image_monitor_ptr(cur_image_ptr); 
    Inc_param(Image_monitor_ptr(cur_image_ptr), REC_CALLS); 
    /* Possiblu changing the cache level */
    if (Use_memoS() && (Getf_memoS(LEVEL_CACHE))) Inc_level_memoS(); 
    result = Test_dp_sat_rec(new_image_ptr, dag_ptr, to_test_lsList_ptr,  
			     Test_K_consistency);
    /* Possibly restoring the cache level */
    if (Use_memoS() && (Getf_memoS(LEVEL_CACHE))) Dec_level_memoS(); 
    /* Get memory allocation data */
    if ((Get_param(Image_monitor_ptr(cur_image_ptr), MAX_SATO_MEMORY) +  
	 Memory_count(new_image_ptr)) > max_mem) 
      max_mem =  
	Get_param(Image_monitor_ptr(cur_image_ptr), MAX_SATO_MEMORY) +  
	Memory_count(new_image_ptr); 
    
    Clear_image(new_image_ptr); 
    
    /* Cleaning up temporary allocations */
    list_status = lsDestroy(*to_test_lsList_ptr, NULL);
    FREE(to_test_lsList_ptr);
#ifdef TRACE_CONS
    if (flag_display) {
      printf("Which is %s\n\n\n", (result ? "consistent" : "unconsistent"));
    }
#endif
    
    /* Memoizing */
    /* Store a satisfiability result */
    if (Use_memoS() && result) {
      Store_memoS(new_alpha_lsList_ptr, beta_vertex_ptr);
      /* After the first storage, try to be compact */
      Setf_memoS(COMPACT_STORAGE);
    }
    /* Eventually store a dependency */
    if (Use_memoS() && !result && Getf_memoS(GET_DEPENDENCIES)) 
      Store_depS(new_alpha_lsList_ptr, beta_vertex_ptr);

    /* Unconsistent? */
    if (!result) {
      /* Monitor */
      Inc_param(Image_monitor_ptr(cur_image_ptr), MODFAIL_NUM); 
      lsFinish(beta_lsGen);
      break; /* break the loop on betas */
    }
  } /* end of the loop on betas */
  
  /* Cleaning up temporary allocations */
  list_status = lsDestroy(*alphas_formula_lsList_ptr, NULL);
  FREE(alphas_formula_lsList_ptr);

  /* Memoizing */
  /* Allow memoizing to create new entries */
  if (Use_memoS()) Clearf_memoS(COMPACT_STORAGE);
  
  /* Monitor */
  if (max_mem > saved_max) 
    Get_param(Image_monitor_ptr(cur_image_ptr), MAX_SATO_MEMORY) = max_mem; 
  else
    Get_param(Image_monitor_ptr(cur_image_ptr), MAX_SATO_MEMORY) = saved_max; 

  /* Consistent? */
  if (result) 
    return NIL(vertex_t);
  else 
    return beta_vertex_ptr;
}

/* ****************************** Plain consistency test */
/* Eventually fires backjumping */
int Plain_test_K_consistency(image_t *cur_image_ptr, dag_t *dag_ptr)
{
  int          flag_display;
  st_table     *agents_st_table_ptr;
  lsList       *alpha_lsList_ptr, *beta_lsList_ptr, *s_formula_lsList_ptr;
  st_generator *at_st_generator;
  stGeneric    ak_stGeneric, av_stGeneric;
  vertex_t     *beta_vertex_ptr;

  /* This is the classification for agents */
  agents_st_table_ptr = st_init_table(st_numcmp, st_numhash);
  /* First of all classify alphas and betas for agents */
  Classify_agents(cur_image_ptr, dag_ptr, agents_st_table_ptr);
  
#ifdef TRACE_CONS
  if (st_count(agents_st_table_ptr) == 0) flag_display = 0;
  if (flag_display) {
    /* Verifying the assignment */
    s_formula_lsList_ptr = Assign_cnf(Main_formula_lsList_ptr(cur_image_ptr),
				      (int*) Value(cur_image_ptr));
    if (lsLength(*s_formula_lsList_ptr) != 0) {
      printf("------------------------------");
      printf("The main formula simplifies to:\n");
      Print_cnf(s_formula_lsList_ptr);
    }
    Free_cnf(s_formula_lsList_ptr);
    printf("------------------------------");
    printf("Current stack:\n");
    print_stack(cur_image_ptr);
    printf("\n");
  }
#else
  s_formula_lsList_ptr = NIL(lsList);
  flag_display = 0;
#endif

#ifdef TRACE_CONS 
  if (st_count(agents_st_table_ptr) == 0) { 
    printf("------------------------------The model is propositional\n\n"); 
  } 
#endif 
  /* Now iterate over agents */
  beta_vertex_ptr = NIL(vertex_t);
  st_foreach_item(agents_st_table_ptr, at_st_generator, &ak_stGeneric, &av_stGeneric) {
    /* These are the current agent alphas and betas */
    alpha_lsList_ptr = Alpha_lsList_ptr(UNCAST_AGENT(av_stGeneric));
    beta_lsList_ptr = Beta_lsList_ptr(UNCAST_AGENT(av_stGeneric));
#ifdef TRACE_CONS
    if (flag_display) {
      if (lsLength(*beta_lsList_ptr) == 0) {
	printf("------------------------------There is no -[]beta in the assigment\n\n");
      }
    }
#endif

    /* If non NIL(vertex_t), beta_vertex_ptr generated an unconsistency */
#ifdef TRACE_CONS
    beta_vertex_ptr = Int_test_K_consistency(cur_image_ptr, dag_ptr, 
					     alpha_lsList_ptr, beta_lsList_ptr, flag_display);
#else
    beta_vertex_ptr = Int_test_K_consistency(cur_image_ptr, dag_ptr, 
					     alpha_lsList_ptr, beta_lsList_ptr, 0);
#endif

    /* Unconsistent? */
    if (beta_vertex_ptr != NIL(vertex_t)) {
      /* Backjumping with strategy1: set the jump point after the final consistency check */
      if (BACK_JUMPING) 
	Set_K_jump_point(cur_image_ptr, dag_ptr, alpha_lsList_ptr, beta_vertex_ptr);
      /* clean the generator for the agents */
      st_free_gen(at_st_generator);
      break; /* break the loop on agents */
    }
  } /* end of the loop on agents */

  /* Cleaning up the agents table */
  st_foreach_item(agents_st_table_ptr, at_st_generator, &ak_stGeneric, &av_stGeneric) {
    Clear_agent(av_stGeneric);
  }
  st_free_table(agents_st_table_ptr);    
#ifdef TRACE_CONS
  if (flag_display) {
    if (beta_vertex_ptr == NIL(vertex_t))
      printf("------------------------------The model is consistent in K\n\n");
    else
      printf("------------------------------The model is not consistent in K\n\n");
  }
#endif
  
  return (beta_vertex_ptr == NIL(vertex_t));
}

/* ****************************** The interface function */
int Test_K_consistency(image_t *cur_image_ptr, dag_t *dag_ptr)
{
  if (EARLY_JUMPING) 
    return Jumpy_test_K_consistency(cur_image_ptr, dag_ptr);
  else
    return Plain_test_K_consistency(cur_image_ptr, dag_ptr);  
}


/* *********************************************************************** */
/* Modal backtracking in K */
/* *********************************************************************** */

/* Perform a smart early pruning (the minimum number of tests is done) */
int Jumpy_test_K_consistency(image_t *cur_image_ptr, dag_t *dag_ptr)
{
  int          first_modal_atom, last_modal_atom;
  int          i, cur_atom, ag_num;
  int          result;
  lsList       *old_beta_lsList_ptr, *old_alpha_lsList_ptr;
  lsList       *new_alpha_lsList_ptr, *new_beta_lsList_ptr;
  vertex_t     *cur_vertex_ptr;
  st_table     *agents_st_table_ptr;
  st_generator *at_st_generator;
  stGeneric    ak_stGeneric, av_stGeneric;

  first_modal_atom = SPREAD(Lpa(dag_ptr) + 1);
  last_modal_atom = SPREAD(Lma(dag_ptr));

  /* This is the classification for agents */
  agents_st_table_ptr = st_init_table(st_numcmp, st_numhash);
  Classify_agents_light(cur_image_ptr, dag_ptr, agents_st_table_ptr);

  old_alpha_lsList_ptr = ALLOC(lsList, 1); old_beta_lsList_ptr = ALLOC(lsList, 1);
  new_alpha_lsList_ptr = ALLOC(lsList, 1); new_beta_lsList_ptr = ALLOC(lsList, 1);

  /* Iterate over agents */
  result = 1;
  st_foreach_item(agents_st_table_ptr, at_st_generator, &ak_stGeneric, &av_stGeneric) {
    ag_num = UNCAST_KEY(ak_stGeneric);

    /* Create old and new lists for each agent */
    *old_alpha_lsList_ptr = lsCreate(); *old_beta_lsList_ptr = lsCreate();
    *new_alpha_lsList_ptr = lsCreate(); *new_beta_lsList_ptr = lsCreate();  

    i = 0;
    /* If jumping already happened, we might want to skip some tests */
    if (Last_stack_idx(cur_image_ptr) > 0) 
      while (i < (Last_stack_idx(cur_image_ptr))) {
	/* Include all the significant alphas and betas into the "old" lists */
	/* Look for a relevant modal atom (must be the same agent) */
	IF_RELEVANT_MODAL_ATOM
	  /* Store the atom in the appropriate list */
	  if (Value(cur_image_ptr)[cur_atom] == TT) 
	    lsNewEnd(*old_alpha_lsList_ptr, CAST_VERTEX_ITEM(cur_vertex_ptr), LS_NH);
	  else 
	    lsNewEnd(*old_beta_lsList_ptr, CAST_VERTEX_ITEM(cur_vertex_ptr), LS_NH);
	END_IF    
	++i;
      }

    /* Look for an early inconsistency */
    /* Running through the stack (up to the end) */
    while (i < Stack_idx(cur_image_ptr)) {
      /* Get the current atom in the stack */
      /* Look for a relevant modal atom (must be the same agent) */
      IF_RELEVANT_MODAL_ATOM
	/* Store the atom in the appropriate list */
	if (Value(cur_image_ptr)[cur_atom] == TT) 
	  lsNewEnd(*new_alpha_lsList_ptr, CAST_VERTEX_ITEM(cur_vertex_ptr), LS_NH);
	else 
	  lsNewEnd(*new_beta_lsList_ptr, CAST_VERTEX_ITEM(cur_vertex_ptr), LS_NH);
      END_IF
      ++i;
    }

    /* See what happened to the lists of alphas and betas */
    /* If there are no new alphas and betas, just go on! */
    if ((lsLength(*new_beta_lsList_ptr) != 0) || 
	(lsLength(*new_alpha_lsList_ptr) != 0)) {

      if (lsLength(*new_beta_lsList_ptr) == 0) {
	if (lsLength(*new_alpha_lsList_ptr) != 0) 
	  /* there were no new betas but some new alphas */
	  Merge_formulae(old_alpha_lsList_ptr, new_alpha_lsList_ptr);
      } else {
	if (lsLength(*new_alpha_lsList_ptr) == 0) 
	  /* There were some betas but no alphas: do not test the old betas again */
	  Swap_formulae(&old_beta_lsList_ptr, &new_beta_lsList_ptr);
	else {
	  /* Everything changed: test everything */
	  Merge_formulae(old_alpha_lsList_ptr, new_alpha_lsList_ptr);
	  Merge_formulae(old_beta_lsList_ptr, new_beta_lsList_ptr);
	}
      }
      
      /* The consistency check */
      result =
	(Int_test_K_consistency(cur_image_ptr, dag_ptr, 
				old_alpha_lsList_ptr, old_beta_lsList_ptr, 0) == NIL(vertex_t));
    }

    /* Clean the lists for the next agent */
    lsDestroy(*old_alpha_lsList_ptr, NULL); lsDestroy(*old_beta_lsList_ptr, NULL);
    lsDestroy(*new_alpha_lsList_ptr, NULL); lsDestroy(*new_beta_lsList_ptr, NULL);

    if (!result) {
      st_free_gen(at_st_generator);
      break; /* break the loop on agents */
    }
  } /* end of the loop on agents */

  /* Clean the allocations */
  FREE(old_alpha_lsList_ptr); FREE(new_alpha_lsList_ptr);
  FREE(old_beta_lsList_ptr); FREE(new_beta_lsList_ptr);
  st_free_table(agents_st_table_ptr);    

  if (result)
    /* Declare consistent the current stack */
    Last_stack_idx(cur_image_ptr) = Stack_idx(cur_image_ptr);
  else
    /* Set an immediate jump point */
    Jump_point(cur_image_ptr) = Backup_idx(cur_image_ptr);

  return result;
}


/* Backjumping: comes into play at the end of the 
   search knowing that the final consistency check generated an inconsistency */

void Set_K_jump_point1(image_t *cur_image_ptr, dag_t *dag_ptr, 
		       lsList *alpha_lsList_ptr, vertex_t *beta_vertex_ptr);
void Set_K_jump_point2(image_t *cur_image_ptr, dag_t *dag_ptr, 
		       lsList *alpha_lsList_ptr, vertex_t *beta_vertex_ptr);

/* EXPERIMENTAL!!! */
void Set_K_jump_point(image_t *cur_image_ptr, dag_t *dag_ptr, 
		      lsList *alpha_lsList_ptr, vertex_t *beta_vertex_ptr)
{
  if (BACK_JUMPING == 1) {
    Set_K_jump_point1(cur_image_ptr, dag_ptr, 
		      alpha_lsList_ptr, beta_vertex_ptr);
  } else if (BACK_JUMPING == 2) {
    Set_K_jump_point2(cur_image_ptr, dag_ptr, 
		      alpha_lsList_ptr, beta_vertex_ptr);
  }
  return;
}

void Set_K_jump_point1(image_t *cur_image_ptr, dag_t *dag_ptr, 
		       lsList *alpha_lsList_ptr, vertex_t *beta_vertex_ptr)
{
  int      first_modal_atom, last_modal_atom;
  int      i, cur_atom;
  int      result, might_jump; 
  lsList   *old_beta_lsList_ptr, *old_alpha_lsList_ptr;
  lsList   *new_alpha_lsList_ptr, *new_beta_lsList_ptr;
  vertex_t *cur_vertex_ptr;
  lsStatus list_status;
  
  first_modal_atom = SPREAD(Lpa(dag_ptr) + 1);
  last_modal_atom = SPREAD(Lma(dag_ptr));

  /* Initialize "old" lists */
  old_alpha_lsList_ptr = ALLOC(lsList, 1);
  *old_alpha_lsList_ptr = lsCreate();
  old_beta_lsList_ptr = ALLOC(lsList, 1);
  *old_beta_lsList_ptr = lsCreate();

  /* Initialize "new" lists */
  new_alpha_lsList_ptr = ALLOC(lsList, 1);
  *new_alpha_lsList_ptr = lsCreate();
  new_beta_lsList_ptr = ALLOC(lsList, 1);
  *new_beta_lsList_ptr = lsCreate();  

  i = 0;
  /* If backjumping already happened, we might want to skip some tests */
  if (Last_stack_idx(cur_image_ptr) > 0) 
    while (i < (Last_stack_idx(cur_image_ptr))) {
      cur_atom = UNVALUATE(Stack(cur_image_ptr)[i].v);
      /* Look for a relevant modal atom (must be the same agent) */
      if (Test_relevance(Stack(cur_image_ptr), Value(cur_image_ptr), i) &&
	  Test_bound(cur_atom, first_modal_atom, last_modal_atom)) {
	cur_vertex_ptr = Xref_vertex_ptr_table(dag_ptr)[UNSPREAD(cur_atom)];
	if (V_value(V_ini_vertex_ptr(cur_vertex_ptr)) == 
	    V_value(V_ini_vertex_ptr(beta_vertex_ptr))) {
	  
	  /* Store the atom in the appropriate list */
	  if (Value(cur_image_ptr)[cur_atom] == TT) 
	    list_status = lsNewEnd(*old_alpha_lsList_ptr, CAST_VERTEX_ITEM(cur_vertex_ptr), LS_NH);
	  else 
	    list_status = lsNewEnd(*old_beta_lsList_ptr, CAST_VERTEX_ITEM(cur_vertex_ptr), LS_NH);

	  ++i;
	}
      }
    }

  result = 1; might_jump = 0;
  /* Look for an early inconsistency */
  while (result) {
    /* Run through the stack (up to the end) */
    while (i < Stack_idx(cur_image_ptr)) {
      /* Get the current atom in the stack */
      cur_atom = UNVALUATE(Stack(cur_image_ptr)[i].v);
      
      /* Look for a relevant modal atom (must be the same agent) */
      if (Test_relevance(Stack(cur_image_ptr), Value(cur_image_ptr), i) &&
	  Test_bound(cur_atom, first_modal_atom, last_modal_atom)) {
	cur_vertex_ptr = Xref_vertex_ptr_table(dag_ptr)[UNSPREAD(cur_atom)];
	if (V_value(V_ini_vertex_ptr(cur_vertex_ptr)) == 
	    V_value(V_ini_vertex_ptr(beta_vertex_ptr))) {

	  /* Store the atom in the appropriate list */
	  if (Value(cur_image_ptr)[cur_atom] == TT) 
	    list_status = lsNewEnd(*new_alpha_lsList_ptr, CAST_VERTEX_ITEM(cur_vertex_ptr), LS_NH);
	  else 
	    list_status = lsNewEnd(*new_beta_lsList_ptr, CAST_VERTEX_ITEM(cur_vertex_ptr), LS_NH);

	  /* It this is a split, it might be the one where to backjump to */
	  if (Stack(cur_image_ptr)[i].f >= SP) might_jump = i;
	}
      }

      /* Now look ahead in the stack and decide what to do */
      ++i;
      if (i < Stack_idx(cur_image_ptr)) {
	cur_atom = UNVALUATE(Stack(cur_image_ptr)[i].v);
	
	/* Look for a modal atom (must be the same agent) */
	if (Test_relevance(Stack(cur_image_ptr), Value(cur_image_ptr), i) &&
	    Test_bound(cur_atom, first_modal_atom, last_modal_atom)) {
	  cur_vertex_ptr = Xref_vertex_ptr_table(dag_ptr)[UNSPREAD(cur_atom)];
	  if (V_value(V_ini_vertex_ptr(cur_vertex_ptr)) == 
	      V_value(V_ini_vertex_ptr(beta_vertex_ptr))) 
	    /* If it is a split then break and try to check */
	    /* In any other case the loop goes on */
	    if (Stack(cur_image_ptr)[i].f >= SP) break;
	}
      }
    }

    /* If we did not reach the end of the stack,
       see what happened to the lists of alphas and betas */
    if (i == Stack_idx(cur_image_ptr)) break;

    if (lsLength(*new_beta_lsList_ptr) == 0) {
      if (lsLength(*new_alpha_lsList_ptr) == 0) {
	/* there where no new alphas and betas */
	/* but, since we are here, there is still stack to go */
	continue; 
      } else {
	/* there were no new betas but some new alphas */
	Merge_formulae(old_alpha_lsList_ptr, new_alpha_lsList_ptr);
	result =
	  (Int_test_K_consistency(cur_image_ptr, dag_ptr, 
				  old_alpha_lsList_ptr, old_beta_lsList_ptr, 0) == NIL(vertex_t));
	/* Reset new_alpha list */
	if (result) {
	  list_status = lsDestroy(*new_alpha_lsList_ptr, NULL);
	  *new_alpha_lsList_ptr = lsCreate();
	}
      }
    } else {
      if (lsLength(*new_alpha_lsList_ptr) == 0) {
	/* There were some betas but no alphas */
	/* do not test the old betas again */
	result =
	  (Int_test_K_consistency(cur_image_ptr, dag_ptr, 
				  old_alpha_lsList_ptr, new_beta_lsList_ptr, 0) == NIL(vertex_t));
	/* Update old betas list and reset the new betas */
	if (result) {
	  Merge_formulae(old_beta_lsList_ptr, new_beta_lsList_ptr);
	  list_status = lsDestroy(*new_beta_lsList_ptr, NULL);
	  *new_beta_lsList_ptr = lsCreate();
	}
      } else {
	/* Everything changed */
	Merge_formulae(old_alpha_lsList_ptr, new_alpha_lsList_ptr);
	Merge_formulae(old_beta_lsList_ptr, new_beta_lsList_ptr);
	result =
	  (Int_test_K_consistency(cur_image_ptr, dag_ptr, 
				  old_alpha_lsList_ptr, old_beta_lsList_ptr, 0) == NIL(vertex_t));
	/* Reset new betas and new alphas */
	if (result) {
	  list_status = lsDestroy(*new_alpha_lsList_ptr, NULL);
	  *new_alpha_lsList_ptr = lsCreate();
	  list_status = lsDestroy(*new_beta_lsList_ptr, NULL);
	  *new_beta_lsList_ptr = lsCreate();
	}
      }
    }
  }
  
  /* Clean the space */
  list_status = lsDestroy(*old_alpha_lsList_ptr, NULL);
  list_status = lsDestroy(*old_beta_lsList_ptr, NULL);
  list_status = lsDestroy(*new_alpha_lsList_ptr, NULL);
  list_status = lsDestroy(*new_beta_lsList_ptr, NULL);
  FREE(old_alpha_lsList_ptr); FREE(new_alpha_lsList_ptr);
  FREE(old_beta_lsList_ptr); FREE(new_beta_lsList_ptr);

  /* Mark up to where tests were already performed */
  Last_stack_idx(cur_image_ptr) = might_jump;

  /* If a backup point was set, let's search in Backup the exact 
     point where to backjump to */
  if ((i < Stack_idx(cur_image_ptr)) && might_jump) {
    cur_atom = UNVALUATE(Stack(cur_image_ptr)[might_jump].v);
    /* Set the jump point */
    for (i = 0; ;i++) 
      if (cur_atom == UNVALUATE(Backup(cur_image_ptr)[i])) break;
    Jump_point(cur_image_ptr) = i + 1;
  }

  return;
}


void Set_K_jump_point2(image_t *cur_image_ptr, dag_t *dag_ptr, 
		       lsList *alpha_lsList_ptr, vertex_t *beta_vertex_ptr)
{
  vertex_t   *alpha_vertex_ptr;
  lsGeneric  t_lsGeneric;
  int i, cur_atom;

  lsLastItem(*alpha_lsList_ptr, &t_lsGeneric, LS_NH);
  alpha_vertex_ptr = UNCAST_VERTEX_ITEM(t_lsGeneric);
  cur_atom = UNVALUATE(Stack(cur_image_ptr)[Stack_idx(cur_image_ptr) - 1].v);

  /* Look back through the stack to find backjump point */
  i = Stack_idx(cur_image_ptr) - 1;
  while (i >= 0) {
    cur_atom = UNVALUATE(Stack(cur_image_ptr)[i].v);
    if ((V_value(alpha_vertex_ptr) == UNSPREAD(cur_atom)) ||
	(V_value(beta_vertex_ptr) == UNSPREAD(cur_atom)))
      break;
    else ++i;
  }
  /* Set the jump point */
  for (i = 0; ;i++) 
    if (cur_atom == UNVALUATE(Backup(cur_image_ptr)[i])) break;
  Jump_point(cur_image_ptr) = i + 1;

  return;
}
  

/* *********************************************************************** */
/* EXPERIMENTS */
/* *********************************************************************** */

/* Prototype of the dicotomic search function */
/*  int Rec_find_jump_point(image_t *cur_image_ptr, dag_t *dag_ptr, int cur_ag, */
/*  			int start_idx, int end_idx, int beta_idx,  */
/*  			int split, int replace); */

/* Backjumping with strategy 2: search dicotomically in the stack for a jump point
   without knowing if the final consistency check generated an inconsistency */
/* Returns 0 if an inconsistency was found, 1 otherwise */
/*  int Set_K_jump_point2(image_t *cur_image_ptr, dag_t *dag_ptr, int cur_ag) */
/*  { */
/*    int         i, cur_atom; */
/*    int         jump_point;  */
  
  /* Explore dicotomically the the stack looking for inconsistency */
  /* Begin from i, end before new_stack_idx, check betas from 0, do not split at once and 
     do not touch old reasons. */
/*    jump_point = Rec_find_jump_point(cur_image_ptr, dag_ptr, cur_ag,  */
/*  				   Last_stack_idx(cur_image_ptr), Stack_idx(cur_image_ptr), 0, 0, 0); */

  /* Go back to the latest split */
/*    if (jump_point != NO_JUMP) { */
/*      while (Stack(cur_image_ptr)[jump_point].f == UP) --jump_point; */
    /* Get the atom corresponding to the jump point */
/*      cur_atom = UNVALUATE(Stack(cur_image_ptr)[jump_point].v); */
    /* The assignment up to the last split is consistent */
/*      Last_stack_idx(cur_image_ptr) = jump_point; */
    /* Set the jump point */
/*      for (i = 0; ;i++)  */
/*        if (cur_atom == UNVALUATE(Backup(cur_image_ptr)[i])) break; */
/*      Jump_point(cur_image_ptr) = i + 1; */
/*    } */
  
/*    return (jump_point == NO_JUMP); */
/*  } */

/*  int Rec_find_jump_point(image_t *cur_image_ptr, dag_t *dag_ptr, int cur_ag,  */
/*  			int start_idx, int end_idx, int beta_idx,  */
/*  			int split, int replace) */
/*  { */
/*    int       i, j, cur_atom, result; */
/*    int       first_modal_atom, last_modal_atom; */
/*    lsList    *new_beta_lsList_ptr, *new_alpha_lsList_ptr; */
/*    vertex_t  *cur_vertex_ptr; */
/*    lsStatus  list_status; */
/*    int       jump_point, pivot;  */
/*    lsGeneric t_lsGeneric; */

/*    first_modal_atom = SPREAD(Lpa(dag_ptr) + 1); */
/*    last_modal_atom = SPREAD(Lma(dag_ptr)); */

  /* Initialize "new" lists */
/*    new_alpha_lsList_ptr = ALLOC(lsList, 1); */
/*    new_beta_lsList_ptr = ALLOC(lsList, 1); */
/*    *new_beta_lsList_ptr = lsCreate();   */
	  
  /* Collect the alphas */
/*    *new_alpha_lsList_ptr = lsCreate(); */
/*    for (j = 0; j < end_idx; j ++) { */
/*      cur_atom = UNVALUATE(Stack(cur_image_ptr)[j].v); */
    /* Look for a relevant modal atom (must be the same agent) */
/*      if (Test_relevance(Stack(cur_image_ptr), Value(cur_image_ptr), j) && */
/*  	Test_bound(cur_atom, first_modal_atom, last_modal_atom)) { */
/*        cur_vertex_ptr = Xref_vertex_ptr_table(dag_ptr)[UNSPREAD(cur_atom)]; */
/*        if ((V_value(V_ini_vertex_ptr(cur_vertex_ptr)) == cur_ag) && */
/*  	  (Value(cur_image_ptr)[cur_atom] == TT))  */
/*  	list_status = lsNewEnd(*new_alpha_lsList_ptr, CAST_VERTEX_ITEM(cur_vertex_ptr), LS_NH); */
/*      } */
/*    } */

/*    if (!split) { */
    /* Check for inconsistency when the upper level does not already provide a beta */
/*      while ((beta_idx < end_idx) && result) { */
      /* Look for the next beta */
/*        cur_atom = UNVALUATE(Stack(cur_image_ptr)[beta_idx].v); */
/*        if (Test_relevance(Stack(cur_image_ptr), Value(cur_image_ptr), beta_idx) && */
/*  	  Test_bound(cur_atom, first_modal_atom, last_modal_atom)) { */
/*  	cur_vertex_ptr = Xref_vertex_ptr_table(dag_ptr)[UNSPREAD(cur_atom)]; */
/*  	if ((V_value(V_ini_vertex_ptr(cur_vertex_ptr)) == cur_ag) && */
/*  	    (Value(cur_image_ptr)[cur_atom] == FF)) { */
	  /* Test consistency */
/*  	  list_status = lsNewEnd(*new_beta_lsList_ptr, CAST_VERTEX_ITEM(cur_vertex_ptr), LS_NH); */
/*  	  result = */
/*  	    (Int_test_K_consistency(cur_image_ptr, dag_ptr,  */
/*  				    new_alpha_lsList_ptr, new_beta_lsList_ptr, 0) == NIL(vertex_t)); */
/*  	  list_status = lsDelEnd(*new_beta_lsList_ptr, &t_lsGeneric); */
	  
/*  	} */
/*        } */
/*        if (result) ++beta_idx; */
/*      } */
/*    } */
  
/*    if (beta_idx == end_idx)  */
    /* All the atoms where scanned without reaching inconsistency */
/*      jump_point = NO_JUMP; */
/*    else { */
    /* There was an inconsistency detected here or in within an higher level */
/*      pivot = start_idx + (end_idx - start_idx) / 2; */
/*      jump_point = NO_JUMP; */
    
    /* Explore the first half of the stack chunk looking for inconsistency
       only if the beta that generated the inconsistency was in the first half */
/*      if ((beta_idx < pivot) && (start_idx != pivot)) */
/*        jump_point = Rec_find_jump_point(cur_image_ptr, dag_ptr, cur_ag, */
/*  				       start_idx, pivot, beta_idx, 0, 1); */
    
/*      if ((jump_point == NO_JUMP) && (start_idx != pivot))  */
/*        jump_point = Rec_find_jump_point(cur_image_ptr, dag_ptr, cur_ag, */
/*  				       pivot, end_idx, beta_idx, 1, 1); */

    /* Failure on the second half means that the correct jump point is latest alpha */
/*      if (jump_point == NO_JUMP) jump_point = end_idx - 1; */
/*    }   */

/*    list_status = lsDestroy(*new_alpha_lsList_ptr, NULL); */
/*    list_status = lsDestroy(*new_beta_lsList_ptr, NULL); */
/*    FREE(new_beta_lsList_ptr); FREE(new_alpha_lsList_ptr); */

/*    return jump_point; */
/*  } */
