/*  *********************************************************************** */
/*  *********************************************************************** */

/*  FILE set.c - *SAT 1.2 */
/*  Memoizing with sets of sets */

/*  *********************************************************************** */
/*  *********************************************************************** */

#include "set.h"

/* Memoizing */
static memo_t *SET_memo_ptr;

/* *********************************************************************** */
/* Initialize and clear memoizing */
/* *********************************************************************** */

/* The function returns a pointer to a single modal level cache */
/* For efficiency reasons we assume that dependencies are a fraction
   DEP_FACTOR of sets */
int_memo_t *Int_init_memoS(dim)
{
  int i;
  int_memo_t *memo_ptr;

  memo_ptr = ALLOC(int_memo_t, 1);
  /* Initialize data and dependencies sets */
  memo_ptr->memo_array = ALLOC(var_set_t*, (2 * dim + 1));
  memo_ptr->dep_array = ALLOC(var_set_t*, (dim + 1));
  memo_ptr->dep_beta = ALLOC(int, (WINDOW_SIZE / DEP_FACTOR));       
  for (i = 0; i <= dim; i++) {
    memo_ptr->memo_array[i] = var_set_new(WINDOW_SIZE);
    memo_ptr->dep_array[i] = var_set_new(WINDOW_SIZE / DEP_FACTOR);
  }
  for (i = dim + 1; i <= (2 * dim); i++) {
    memo_ptr->memo_array[i] = var_set_new(WINDOW_SIZE);
  }
  memo_ptr->cur_set = -1;
  memo_ptr->cur_dep = -1;
  /* Fast growth and no compaction */
  memo_ptr->growth = 0;
  memo_ptr->compact = 0;

  /* Trashing is not started yet */
  memo_ptr->trashing_sets = 0;
  memo_ptr->trashing_deps = 0;

  /* Compute allocated memory (bytes) */
  SET_memo_ptr->stats[MEMORY] += 
    sizeof(memo_t) +                                     /* The structure */
    (2 * SET_memo_ptr->dim *                             /* Sets */
     (sizeof(var_set_t) + WINDOW_SIZE / sizeof(int))) +
    (SET_memo_ptr->dim *                                 /* Dependencies */
     (sizeof(var_set_t) + 
      (WINDOW_SIZE / DEP_FACTOR) / sizeof(int))) +
    sizeof(int) * (WINDOW_SIZE / DEP_FACTOR);            /* Dependency labels */ 

  return memo_ptr;
}

void Init_memoS(int dim)
{
  int i;
  SET_memo_ptr = ALLOC(memo_t, 1);
  /* Initialize array of cache references */
  for (i = 0; i < MAX_DEEP; i++) 
    SET_memo_ptr->memo_of_level[i] = NIL(int_memo_t);

  /* Record the dimensions */
  SET_memo_ptr->dim = dim;

  /* All flags disabled */
  for (i = 0; i < MAX_FLAGS; i++)  
    SET_memo_ptr->flags[i] = 0;

  /* All stats are empty */
  for (i = 0; i < MAX_STATS; i++)  
    SET_memo_ptr->stats[i] = 0;

  /* Default level is 0 */
  SET_memo_ptr->level = 0;

  /* Initialize current level cache */
  SET_memo_ptr->memo_of_level[0] = Int_init_memoS(dim);
  
  return;
}

void Int_clear_memoS(int_memo_t *memo_ptr, int dim)
{
  return;
}

void Clear_memoS()
{
  int i = 0;
  /* Clean every allocated cache */
  while (SET_memo_ptr->memo_of_level[i] != NIL(int_memo_t)) {
    Int_clear_memoS(SET_memo_ptr->memo_of_level[i], SET_memo_ptr->dim);
    ++i;
  }
  FREE(SET_memo_ptr);
  return;
}

void Force_no_memoS()
{
  SET_memo_ptr = NIL(memo_t);
  return;
}


/* *********************************************************************** */
/* Accessing the structure */
/* *********************************************************************** */

int Use_memoS()
{
  return (SET_memo_ptr != NIL(memo_t));
}

void Setf_memoS(int flag)
{
  if (flag == COMPACT_STORAGE)
    SET_memo_ptr->memo_of_level[SET_memo_ptr->level]->compact = 1;
  SET_memo_ptr->flags[flag] = 1;
  return;
}

void Clearf_memoS(int flag)
{
  if (flag == COMPACT_STORAGE)
    SET_memo_ptr->memo_of_level[SET_memo_ptr->level]->compact = 0;
  SET_memo_ptr->flags[flag] = 0;
  return;
}

int Getf_memoS(int flag)
{
  if (flag == COMPACT_STORAGE)
    return SET_memo_ptr->memo_of_level[SET_memo_ptr->level]->compact;
  return SET_memo_ptr->flags[flag];
}

int Stats_memoS(int stat)
{
  return SET_memo_ptr->stats[stat];
}

void Inc_level_memoS()
{
  ++(SET_memo_ptr->level);
}

void Dec_level_memoS()
{
  --(SET_memo_ptr->level);
}

int Get_level_memoS()
{
  return SET_memo_ptr->level;
}

lsList *Test_memoS(lsList *alpha_lsList_ptr, lsList *beta_lsList_ptr)
{
  var_set_t  *witness_var_set_ptr, *betas_var_set_ptr;
  lsGen      t_lsGen;
  lsGeneric  vertex_lsGeneric;
  lsStatus   list_status;
  int        alpha_value, beta_value, i, j;
  int_memo_t *memo_ptr;

  /* First check if the cache for the current level is there 
     and if it is not there then create it */
  memo_ptr = SET_memo_ptr->memo_of_level[SET_memo_ptr->level];
  if (memo_ptr == NIL(int_memo_t)) {
    SET_memo_ptr->memo_of_level[SET_memo_ptr->level] = 
      Int_init_memoS(SET_memo_ptr->dim);
    memo_ptr = SET_memo_ptr->memo_of_level[SET_memo_ptr->level];
  }

  if (memo_ptr->cur_set != -1) {
    /* Create a witness, fill it with all the elements */
    witness_var_set_ptr = var_set_new(WINDOW_SIZE);
    for (i = 0; i < WINDOW_SIZE; i++) 
      var_set_set_elt(witness_var_set_ptr, i);
    
    /* Filter the witness through the alphas */
    lsForEachItem(*alpha_lsList_ptr, t_lsGen, vertex_lsGeneric) {
      ++(SET_memo_ptr->stats[MEMO_ACCESS]);
      alpha_value = V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric));
      var_set_and(witness_var_set_ptr, 
		  witness_var_set_ptr, memo_ptr->memo_array[alpha_value]);
    }
    
    /* If at least one element of the witness is set, go on 
       otherwise fail */
    if (var_set_is_empty(witness_var_set_ptr)) {
      var_set_free(witness_var_set_ptr);
      return beta_lsList_ptr;
    }

    /* Transfer the list of betas to a set */
    betas_var_set_ptr = var_set_new(SET_memo_ptr->dim + 1);
    lsForEachItem(*beta_lsList_ptr, t_lsGen, vertex_lsGeneric) {
      beta_value = V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric));
      var_set_set_elt(betas_var_set_ptr, beta_value);
    }

    /* Now try to discard as many betas as possible */
    i = 0;
    while ((i < WINDOW_SIZE) && !var_set_is_empty(betas_var_set_ptr)) {
      if (var_set_get_elt(witness_var_set_ptr, i)) {
	for (j = 1; j <= SET_memo_ptr->dim; j++)
	  if (var_set_get_elt(betas_var_set_ptr, j) &&
	      var_set_get_elt(memo_ptr->memo_array[j + SET_memo_ptr->dim], i)) {
	    ++(SET_memo_ptr->stats[MEMO_SUCCESS]);
	    var_set_clear_elt(betas_var_set_ptr, j);
	  }
      }
      ++i;
    }

    if (var_set_is_empty(betas_var_set_ptr)) {
      /* All the betas were consistent! */
      beta_lsList_ptr = NIL(lsList);
    } else {
      /* Otherwise, just discard betas from the list */
      lsForEachItem(*beta_lsList_ptr, t_lsGen, vertex_lsGeneric) {
	beta_value = V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric));
	if (!var_set_get_elt(betas_var_set_ptr, beta_value)) 
	  list_status = lsDelBefore(t_lsGen, &vertex_lsGeneric);
      }
    }

    /* Destroy temporary allocation */
    var_set_free(witness_var_set_ptr);
    var_set_free(betas_var_set_ptr);

  }
  
  return beta_lsList_ptr;
}

int Trash_subsets(int cur_set);

void Store_memoS(lsList *alpha_lsList_ptr, vertex_t *beta_vertex_ptr)
{
  lsGen      t_lsGen;
  lsGeneric  vertex_lsGeneric;
  int        alpha_value, beta_value, i, cur_set;
  int_memo_t *memo_ptr;

  /* Set current cache */
  /* It must be there since testing occurs always before than storing
     and testing takes care of allocating new caches */
  memo_ptr = SET_memo_ptr->memo_of_level[SET_memo_ptr->level];

  if (!Getf_memoS(COMPACT_STORAGE)) {
    ++(SET_memo_ptr->stats[MEMO_STORAGE]);

    /* Increment the current set modulo WINDOW_SIZE */
    (memo_ptr->cur_set)++;
    if (memo_ptr->cur_set == WINDOW_SIZE) {
      memo_ptr->cur_set = 0;
      /* Growth is now slow */
      if (Getf_memoS(TRASH_SUBSETS)) 
	memo_ptr->growth = 1;
      /* Trashing of sets starts */
      memo_ptr->trashing_sets = 1;
    }
    
    /* If slow growth and TRASH_SUBSETS are set then 
       try to trash subsets only */
    if ((Getf_memoS(TRASH_SUBSETS) && (memo_ptr->growth))) {
      cur_set = Trash_subsets(memo_ptr->cur_set);
      /* If no subset to trash was found... */
      if (cur_set == WINDOW_SIZE) 
	/* ...return to fast growth */
	memo_ptr->growth = 0;
      else 
	/* ... else update the current set */
	memo_ptr->cur_set = cur_set;
    } 

    if ((memo_ptr->trashing_sets) && (!(memo_ptr->growth)))
      ++(SET_memo_ptr->stats[MEMO_TRASHING]);

    /* Be sure to clear everything before setting */
    for (i = 0; i <= (2 * SET_memo_ptr->dim); i++) 
      var_set_clear_elt(memo_ptr->memo_array[i],memo_ptr->cur_set);
    
    /* Set alphas properly */
    lsForEachItem(*alpha_lsList_ptr, t_lsGen, vertex_lsGeneric) {
      alpha_value = V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric));
      var_set_set_elt(memo_ptr->memo_array[alpha_value],
		      memo_ptr->cur_set);
    }

  } 
  
  /* Set the beta properly (either for a new entry or for an old one) */
  beta_value = V_value(beta_vertex_ptr) + SET_memo_ptr->dim;
  var_set_set_elt(memo_ptr->memo_array[beta_value],
		  memo_ptr->cur_set);

  return;
}
  

int Trash_subsets(int cur_set)
{
  var_set_t  *witness_var_set_ptr;
  int        *modal_atoms;
  int        max_atoms, i, run_set;
  int_memo_t *memo_ptr;
  
  memo_ptr = SET_memo_ptr->memo_of_level[SET_memo_ptr->level];

  modal_atoms = ALLOC(int, (2 * SET_memo_ptr->dim + 1));
  run_set = cur_set;
  while (run_set < WINDOW_SIZE) {
    max_atoms = 0;
    /* Collect all the atoms of run_set */
    for (i = 0; i <= (2 * SET_memo_ptr->dim); i++) 
      if (var_set_get_elt(memo_ptr->memo_array[i], run_set)) 
	modal_atoms[max_atoms++] = i;

    /* Create a witness, fill it with all the elements but the current */
    witness_var_set_ptr = var_set_new(WINDOW_SIZE);
    for (i = 0; i < WINDOW_SIZE; i++) 
      if (i != run_set) var_set_set_elt(witness_var_set_ptr, i);

    /* Now look for possible supersets in the structure 
       (other than run_set itself!)*/
    for (i = 0; i < max_atoms; i++) 
      var_set_and(witness_var_set_ptr, 
		  witness_var_set_ptr, memo_ptr->memo_array[modal_atoms[i]]);

    /* If the resulting set is not empty, supersets were found
       so run_set can be trashed */
    if (!var_set_is_empty(witness_var_set_ptr)) {
      var_set_free(witness_var_set_ptr);
      break;
    }
	
    /* No supersets found, try with the next set */
    var_set_free(witness_var_set_ptr);
    ++run_set;
  }
      
  FREE(modal_atoms);
  return run_set;
}

lsList *Test_depS(lsList *alpha_lsList_ptr, lsList *beta_lsList_ptr,
		  vertex_t **rbeta_vertex_ptr_ptr)
{
  var_set_t  *witness_var_set_ptr;
  var_set_t  *alphas_var_set_ptr, *betas_var_set_ptr;
  lsGen      t_lsGen;
  lsGeneric  vertex_lsGeneric;
  lsStatus   list_status;
  int        atom_value, i, j;
  int_memo_t *memo_ptr;
  int        win_size;
  int        *purged_alphas, purged_alpha_idx;

  *rbeta_vertex_ptr_ptr = NIL(vertex_t);

  /* First check if the cache for the current level is there 
     and if it is not there then create it */
  memo_ptr = SET_memo_ptr->memo_of_level[SET_memo_ptr->level];
  if (memo_ptr == NIL(int_memo_t)) {
    SET_memo_ptr->memo_of_level[SET_memo_ptr->level] = 
      Int_init_memoS(SET_memo_ptr->dim);
    memo_ptr = SET_memo_ptr->memo_of_level[SET_memo_ptr->level];
  }

  /* Set the reduced window size */
  win_size = WINDOW_SIZE / DEP_FACTOR;

  if (memo_ptr->cur_dep != -1) {
    /* Create a witness, fill it with all the elements */
    witness_var_set_ptr = var_set_new(win_size);
    for (i = 0; i < win_size; i++) 
      var_set_set_elt(witness_var_set_ptr, i);
    
    alphas_var_set_ptr = var_set_new(SET_memo_ptr->dim + 1);
    betas_var_set_ptr = var_set_new(SET_memo_ptr->dim + 1);
    /* Put the alpha and beta lists into sets */
    lsForEachItem(*alpha_lsList_ptr, t_lsGen, vertex_lsGeneric) {
      atom_value = V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric));
      var_set_set_elt(alphas_var_set_ptr, atom_value);
    }
    lsForEachItem(*beta_lsList_ptr, t_lsGen, vertex_lsGeneric) {
      atom_value = V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric));
      var_set_set_elt(betas_var_set_ptr, atom_value);
    }
    /* Filter the witness through the complement of the alphas */
    /* (seeking for cached subsets of the current set of alphas) */
    for (i = 0; i <= SET_memo_ptr->dim; i++) 
      if (!var_set_get_elt(alphas_var_set_ptr, i)) {
	++(SET_memo_ptr->stats[DEP_ACCESS]);
	var_set_and(witness_var_set_ptr, 
		    witness_var_set_ptr, memo_ptr->dep_array[i]);
      }

    /* If at least one element of the witness is set, go on 
       otherwise fail */
    if (var_set_is_empty(witness_var_set_ptr)) {
      var_set_free(alphas_var_set_ptr);
      var_set_free(betas_var_set_ptr);
      var_set_free(witness_var_set_ptr);
      return alpha_lsList_ptr;
    }

#ifdef TRACE_DEP
    printf("The set "); Print_alpha(alpha_lsList_ptr); 
    printf("contains dependencies\n");
#endif

    /* We must keep track of purged alphas */
    purged_alpha_idx = 0;
    purged_alphas = ALLOC(int, (SET_memo_ptr->dim + 1));
    
    /* Now try to discard as many alphas as possible */
    for (i = 0; i < win_size; i++) 
      if (var_set_get_elt(witness_var_set_ptr, i)) {

	/* The set i contains already purged alphas? */
	j = 0;
	while (j < purged_alpha_idx) {
	  if (!var_set_get_elt(memo_ptr->dep_array[purged_alphas[j]],i)) break;
	  ++j;
	}
	/* If so,  it is no more a subset of current alphas */ 
	if (j < purged_alpha_idx) continue;  

	/* Two cases matter: (i) the beta label is contained in the current list
           of alphas or (ii) the beta label is contained in the current list of betas */
	if (var_set_get_elt(alphas_var_set_ptr, memo_ptr->dep_beta[i])) {
	  /* case (i) applies: the alpha corresponding to the beta label is redundant */
	  ++(SET_memo_ptr->stats[DEP_SUCCESS_PRUNE]);
#ifdef TRACE_DEP
	  {
	    int k;
	    printf("From the set ");
	    for (k = 1; k <= SET_memo_ptr->dim; k++) 
	      if (var_set_get_elt(alphas_var_set_ptr, k)) printf("%d ", k);
	    printf("\n %d is extracted because of dependency:\n", memo_ptr->dep_beta[i]);
	    for (k = 1; k <= SET_memo_ptr->dim; k++) 
	      if (!var_set_get_elt(memo_ptr->dep_array[k], i)) printf("%d ", k);
	    printf("-> %d\n", memo_ptr->dep_beta[i]);
	  }
#endif
	  var_set_clear_elt(alphas_var_set_ptr, memo_ptr->dep_beta[i]);
	  purged_alphas[purged_alpha_idx++] = memo_ptr->dep_beta[i];
	} else if (var_set_get_elt(betas_var_set_ptr, memo_ptr->dep_beta[i])) {
	  ++(SET_memo_ptr->stats[DEP_SUCCESS_UNSAT]);
	  /* case (ii) applies: the current test is uncosistent */
	  break;
	}
      }

    /* Clear array of purged alphas */
    FREE(purged_alphas);

    if (i < win_size) {
      /* If the loop ended before reaching winsize an inconsistency was found */
      alpha_lsList_ptr = NIL(lsList);
      lsForEachItem(*beta_lsList_ptr, t_lsGen, vertex_lsGeneric) {
	atom_value = V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric));
	if (atom_value ==  memo_ptr->dep_beta[i]) {
	  *rbeta_vertex_ptr_ptr = UNCAST_VERTEX_ITEM(vertex_lsGeneric);
	  break;
	}
      }
      /* The loop ALWAYS breaks */
      list_status = lsFinish(t_lsGen);
    } else {
      /* Otherwise, just discard alphas from the list */
      lsForEachItem(*alpha_lsList_ptr, t_lsGen, vertex_lsGeneric) {
	atom_value = V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric));
	if (!var_set_get_elt(alphas_var_set_ptr, atom_value)) 
	  list_status = lsDelBefore(t_lsGen, &vertex_lsGeneric);
      }
    }
    
    /* Destroy temporary allocations */
    var_set_free(alphas_var_set_ptr);
    var_set_free(betas_var_set_ptr);
    var_set_free(witness_var_set_ptr);
  }
  
  return alpha_lsList_ptr;
}

void Store_depS(lsList *alpha_lsList_ptr, vertex_t *beta_vertex_ptr)
{
  int        i, alpha_value;
  lsGen      t_lsGen;
  lsGeneric  vertex_lsGeneric;
  int_memo_t *memo_ptr;
  int        win_size;

  /* Set current cache */
  /* It must be there since testing occurs always before than storing
     and testing takes care of allocating new caches */
  memo_ptr = SET_memo_ptr->memo_of_level[SET_memo_ptr->level];
  win_size = WINDOW_SIZE / DEP_FACTOR;

  ++(SET_memo_ptr->stats[DEP_STORAGE]);

  /* Increment the current set modulo win_size */
  (memo_ptr->cur_dep)++;
  if (memo_ptr->cur_dep == win_size) {
    memo_ptr->cur_dep = 0;
    memo_ptr->trashing_deps = 1;
  }

  if (memo_ptr->trashing_deps)
    ++(SET_memo_ptr->stats[DEP_TRASHING]);

  /* Be sure to clear everything */
  for (i = 0; i <= (SET_memo_ptr->dim); i++) 
    var_set_set_elt(memo_ptr->dep_array[i],memo_ptr->cur_dep);
  
#ifdef TRACE_DEP
  printf("The dependency "); Print_alpha_beta(alpha_lsList_ptr, beta_vertex_ptr);
  printf("is stored\n");
#endif

  /* Set alphas properly 
     (the complement of the set of alphas is stored) */
  lsForEachItem(*alpha_lsList_ptr, t_lsGen, vertex_lsGeneric) {
    alpha_value = V_value(UNCAST_VERTEX_ITEM(vertex_lsGeneric));
    var_set_clear_elt(memo_ptr->dep_array[alpha_value],
		      memo_ptr->cur_dep);
  }
  /* Label the dependency set with the beta */
  memo_ptr->dep_beta[memo_ptr->cur_dep] = V_value(beta_vertex_ptr);

  return;
}

  
